/*
 * @(#)Function.java
 *
 * $Id: Function.java,v 1.22 2000/12/06 03:13:18 huaz Exp $
 *
 * Created on October 30, 2000, 10:29 AM
 */

// IMPORTANT **************************************************
/* NOTE: Al must do javadoc -package Function.java */
// IMPORTANT **************************************************

import java.util.*;
import java.io.*;

/**
 * Function classes used only by Formula to evaluate functions.
 * Any function needs to have a function handler that implements
 * the "evaluate" interface of the base class Function.
 *
 * A function can accept zero, one, or more parameters.  Each parameter
 * is a number, a relative/absolute address or an address range (e.g., A1:B3).
 *
 * @author Hua Zhong
 * @version $Revision: 1.22 $
 * @see Node
 * @see UMTableModel
 * @see LinkedList
 * @see Formula
 */
public abstract class Function {

    static private ParserException exception = new ParserException("#PARAM?");

    
    /**
     * Checks whether the specified parameter node is an address range.
     *
     * @param param a Node
     * @return true if only value and the node is of Node.COLON
     */    
    protected boolean isRange(Node param) {
	LinkedList exp = param.getExp();
	return exp.size() == 1 &&
	    ((Node)exp.getFirst()).isType(Node.COLON);
	    //((Node)param.getExp().getFirst()).isType(Node.COLON);
    }

    
    /**
     * Returns the first node of a specified parameter.
     *
     * @param param a Node
     * @return Node gets first value of node
     */    
    protected Node getFirst(Node param) {
	return (Node)param.getExp().getFirst();
    }

    
    /**
     * Checks whether this function has any parameter.
     *
     * @param func a Node
     * @throws ParserException throws an exception which contains the parser
     */    
    protected void checkParamsExist(Node func) throws ParserException {

	if (func.getParams().size()==0)
	    throw exception;
    }

    /**
     * This gets the first float number of a parameter list, for functions
     * only accepting a single parameter such as <code>ABS</code>, <code>COS
     * </code>, etc.
     *
     * @param table the SharpTabelModel
     * @param node the formula unit
     * @param col the int column coordinate
     * @param row the int row coordinate
     * @return the float number
     * @throws ParserException throws an exception which contains the parser
     *
     */
    static protected float getSingleParameter(UMTableModel table, Node node,
					    int row, int col)
	throws ParserException {
	//	Node param = node.getNextParam();
	LinkedList params = node.getParams();

	if (params.size() != 1)
	    throw new ParserException("#PARAM?");

	LinkedList exp = ((Node)params.getFirst()).getExp();

	return Formula.evaluate(table, exp, row, col).floatValue();

    }

    /**
     * This should be implemented in each function.
     *
     * @param table the SharpTabelModel
     * @param node the function node starting with the funciton name
     * with a chain of parameters
     * @param col the int column coordinate
     * @param row the int row coordinate
     * @exception ParserException throws an exception which contains the parser
     * @return Number a value
     */
    public abstract Number evaluate(UMTableModel table, Node node,
				   int row, int col) throws ParserException;

    /**
     *  Return the usage of the function
     *
     *  @return a String
     */
    public abstract String getUsage();

    /**
     * Return the description of the function
     *
     * @return a String
     */
    public abstract String getDescription();

    /**
     * Checks whether this function requires parameters. By default yes.
     * @see FunctionPI
     * @see FunctionE
     * @return always returns true
     */
    public boolean requireParams() { return true; }
}

/**
 * Class: FunctionSum
 *
 * Description: This will sum all of the values in the parameter list
 * from Node. It gets the values from UMTableModel.
 *
 * <p> Here's an example: <br>
 *
 * <code>SUM</code><br>
 *    usage: <code>=SUM(parameter list)</code><br>
 *    returns the arithmetic sum of the specified parameters<br>
 *    example: <code>=SUM(-1,2,57)</code> returns <code>58.0</code>
 */
class FunctionSum extends Function {

    
    /**
     * Sums all the values for the specified row and column
     *
     * @param table the actual values to sum
     * @param node a LinkedList
     * @param col the int column coordinate
     * @param row the int row coordinate
     * @throws ParserException throws an exception which contains the parser
     * @return Number the value calculated
     */    
    public Number evaluate(UMTableModel table, Node node,
			  int row, int col) throws ParserException {
	// requires parameters
	checkParamsExist(node);

	float sum = 0;

	LinkedList params = node.getParams();
	if (params != null) {

	    // go over the parameters
	    Iterator it = params.iterator();

	    while (it.hasNext()) {

		// get this parameter
		Node exp = (Node)it.next();

		// if it's a range of cells
		if (isRange(exp)) {
		    CellPoint[] addr =
			getFirst(exp).getAddressRange(row, col);
		    // for a range, go over the whole range
		    for (int i = addr[0].getRow(); i <= addr[1].getRow(); i++)
			for (int j = addr[0].getCol(); j <= addr[1].getCol(); j++)
			    // get the numeric value of that cell
			    sum += table.getNumericValueAt(i, j).floatValue();
                         Debug.println("Printing Sum " + sum);
		}
		else {
		    // evaluate this parameter's expression (sub-formula)
		    sum += Formula.evaluate(table, exp.getExp(), row, col).floatValue();
		}
	    }

	}

	return (Number)(new Float(sum));
    }

    /**
     * Gets what this function is doing
     *
     * @return a string */    
    public String getUsage() {
	return "SUM(value1,value2,...)";
    }

    /**
     * Gets the description of this function
     *
     * @return a string */    
    public String getDescription() {
	return "Adds all the numbers in a set of values.";
    }
}


/**
 * Class: FunctionCount
 *
 * Description: This will count all of the values in the parameter list
 * from Node. It gets the values from UMTableModel.
 *
 * <p> Here's an example: <br>
 *
 * <code>COUNT</code><br>
 *   usage: <code>=COUNT(parameter list)</code><br>
 *   returns the number of parameters specified<br>
 *   example: <code>=COUNT(A1:A7)</code> returns <code>7.0</code>
 */
class FunctionCount extends Function {

    /**
     * Counts all the values for the specififed row and column
     *
     * @param table the actual values to sum
     * @param node a LinkedList
     * @param col the int column coordinate
     * @param row the int row coordinate
     * @throws ParserException throws an exception which contains the parser
     * @return Number the value calculated
     */  
    public Number evaluate(UMTableModel table, Node node,
			int row, int col) throws ParserException {

	// requires parameters
	checkParamsExist(node);

	int count = 0;

	LinkedList params = node.getParams();

	if (params != null) {

	    Iterator it = params.iterator();

	    while (it.hasNext()) {
		// the first parameter
		Node exp = (Node)it.next();

		if (isRange(exp)) {
		    // if it's a range then count the number of cells
		    CellPoint[] addr =
			getFirst(exp).getAddressRange(row, col);
		    count += (addr[1].getRow()-addr[0].getRow()+1) *
			(addr[1].getCol()-addr[0].getCol()+1);
		}
		else {
		    // otherwise count one
		    count++;
		}
	    }

	}

	return new Integer(count);
    }
    
    
    /**
     * Gets what this function is doing
     *
     * @return a string */    
    public String getUsage() {
	return "COUNT(value1,value2,...)";
    }

    /**
     * Gets the description of this function
     *
     * @return a string*/   
    public String getDescription() {
	return "Counts the number of cells that contain numbers and numbers within the list of arguments.";
    }
}

/**
 * Class: FunctionAverage
 *
 * Description: This will average all of the values in the parameter list
 * from Node. It gets the values from UMTableModel.
 *
 * <p> Here's an example: <br>
 *
 * <code>AVERAGE</code><br>
 *   usage: <code>=AVERAGE(parameter list)</code><br>
 *   returns the arithmetic mean of the specified parameters<br>
 *   example: <code>=AVERAGE(1,2,3)</code> returns <code>2.0</code>
 */
class FunctionAverage extends Function {

    /**
     * Finds the average for the specified row and column
     *
     * @param table the actual values to sum
     * @param node a LinkedList
     * @param col the int column coordinate
     * @param row the int row coordinate
     * @throws ParserException throws an exception which contains the parser
     * @return Number the value calculated
     */    
    public Number evaluate(UMTableModel table, Node node,
			  int row, int col) throws ParserException {
	float sum = (new FunctionSum()).evaluate(table, node, row, col).
	    floatValue();
	float nCells = (new FunctionCount()).evaluate(table, node, row, col).
	    floatValue();
	return new Float(sum/nCells);
    }

    /**
     * Gets what this function is doing
     *
     * @return a string */      
    public String getUsage() {
	return "AVERAGE(value1,value2,...)";
    }

    /**
     * Gets the description of this function
     *
     * @return a string */     
    public String getDescription() {
	return "Returns the average (arithmetric mean) of its arguments.";
    }
}

/**
 * Class: FunctionMin
 *
 * Description: This will find the min of all of the values in the parameter 
 * list from Node. It gets the values from UMTableModel.
 *
 * <p> Here's an example: <br>
 *
 * <code>MIN</code><br>
 *   usage: <code>=MIN(parameter list)</code><br>
 *   returns the minimum value of the specified parameters<br>
 *   example: <code>=MIN(5,6,-1)</code> returns <code>-1.0</code>
 */
class FunctionMin extends Function {

    /**
     * Finds the minimum value for the specififed row and column
     *
     * @param table the actual values to sum
     * @param node a LinkedList
     * @param col the int column coordinate
     * @param row the int row coordinate
     * @throws ParserException throws an exception which contains the parser
     * @return Number the value calculated
     */    
    public Number evaluate(UMTableModel table, Node node,
			  int row, int col) throws ParserException {

        int operation =0;
	// requires parameters
	checkParamsExist(node);

	float min = Float.MAX_VALUE;
	LinkedList params = node.getParams();

	if (params != null) {

	    Iterator it = params.iterator();

	    while (it.hasNext()) {
		Node exp = (Node)it.next();
		if (isRange(exp)) {
		    CellPoint[] addr =
			getFirst(exp).getAddressRange(row, col);
		    for (int i = addr[0].getRow(); i <= addr[1].getRow(); i++)
			for (int j = addr[0].getCol(); j <= addr[1].getCol(); j++){
			    float value =
				table.getNumericValueAt(i, j).floatValue();
			    if (value < min){
                                operation++;
				min = value;
                            }
			}
		}
		else {
		    float value =
			Formula.evaluate(table, exp.getExp(), row, col).floatValue();
		    if (value < min){
                        operation++;
			min = value;
                    }
		}
	    }

	}
        //if(operation !=0)
	 return new Float(min);
      //return new Float(0);
    }

    /**
     * Gets what this function is doing
     *
     * @return a string */    
    public String getUsage() {
	return "MIN(value1,value2,...)";
    }

    /**
     * Gets the description of this function
     *
     * @return a string */    
    public String getDescription() {
	return "Returns the smallest number in a set of values.";
    }
}

/**
 * Class: FunctionMax
 *
 * Description: This will find the max of all of the values in the parameter 
 * list from Node. It gets the values from UMTableModel.
 *
 * <p> Here's an example: <br>
 *
 * <code>MAX</code><br>
 *   usage: <code>=MAX(parameter list)</code><br>
 *   returns the maximum value of the specified parameters<br>
 *   example: <code>=MAX(5,6,-1)</code> returns <code>6.0</code>
 */
class FunctionMax extends Function {


    /**
     *
     * Finds the maximum value for the specififed row and column
     *
     * @param table the actual values to sum
     * @param node a LinkedList
     * @param col the int column coordinate
     * @param row the int row coordinate
     * @throws ParserException throws an exception which contains the parser
     * @return Number the value calculated
     */     
    public Number evaluate(UMTableModel table, Node node,
			  int row, int col) throws ParserException {
        int operation = 0;

	// requires parameters
	checkParamsExist(node);

	float max = Float.MIN_VALUE;
	LinkedList params = node.getParams();

	if (params != null) {

	    Iterator it = params.iterator();

	    while (it.hasNext()) {

		Node exp = (Node)it.next();

		if (isRange(exp)) {
		    CellPoint[] addr =
			getFirst(exp).getAddressRange(row, col);

		    for (int i = addr[0].getRow(); i <= addr[1].getRow(); i++)
			for (int j = addr[0].getCol(); j <= addr[1].getCol(); j++){
			    float value =
				table.getNumericValueAt(i, j).floatValue();
			    if (value > max){
                                operation ++;
				max = value;
                            }
			}
		}
		else {
		    float value =
			Formula.evaluate(table, exp.getExp(), row, col).floatValue();
		    if (value > max){
                        operation++;
			max = value;
                    }
		}
	    }

	}
        if (operation != 0)
	return new Float(max);
        else return new Float(0);
    }


    /**
     * Gets what this function is doing
     *
     * @return a string */      
    public String getUsage() {
	return "MAX(value1,value2,...)";
    }

    /**
     * Gets the description of this function
     *
     * @return a string */     
    public String getDescription() {
	return "Returns the largest number in a set of values.";
    }
}

/**
 * Class: FunctionSin
 *
 * Description: This will find the sine of all of the values in the parameter 
 * list from Node. It gets the values from UMTableModel.
 *
 * <p> Here's an example: <br>
 *
 * <code>SIN</code><br>
 *   usage: <code>=SIN(parameter)</code><br>
 *   accepts only one literal or address<br>
 *   returns the sine of the specified parameter (in radians)<br>
 *   example: <code>=SIN(45)</code> returns <code>0.8509035</code>
 */
class FunctionSin extends Function {

    /**
     * Finds the sine of the value at the specified row and column
     *
     * @param table the actual values to sum
     * @param node a LinkedList
     * @param col the int column coordinate
     * @param row the int row coordinate
     * @throws ParserException throws an exception which contains the parser
     * @return Number the value calculated
     */     
    public Number evaluate(UMTableModel table, Node node,
			  int row, int col) throws ParserException {
	return new Float(Math.sin
			 (getSingleParameter(table, node, row, col)));
    }

    /**
     * Gets what this function is doing
     *
     * @return a string */      
    public String getUsage() {
	return "SIN(value)";
    }

    /**
     * Gets the description of this function
     *
     * @return a string */     
    public String getDescription() {
	return "Returns the sine of an angle.";
    }
}

/**
 * Class: FunctionSum
 *
 * Description: This will find the square root of all of the values in the 
 * parameter list from Node. It gets the values from UMTableModel.
 *
 * <p> Here's an example: <br>
 *
 * <code>SQRT</code><br>
 *   usage: <code>=STDDEV(parameter)</code><br>
 *   returns the standard deviation of the specified parameter<br>
 *   example: <code>=SQRT(19044)</code> returns <code>138.0</code>
 */
class FunctionSqrt extends Function {

    /**
     * Finds the square root of the value at the specified row and column.
     *
     * @param table the actual values to sum
     * @param node a LinkedList
     * @param col the int column coordinate
     * @param row the int row coordinate
     * @throws ParserException throws an exception which contains the parser
     * @return Number the value calculated
     */    
    public Number evaluate(UMTableModel table, Node node,
			  int row, int col) throws ParserException {

	return new Float(Math.sqrt
			 (getSingleParameter(table, node, row, col)));
    }

    /**
     * Gets what this function is doing
     *
     * @return a string */      
    public String getUsage() {
	return "SQRT(value)";
    }

    /**
     * Gets the description of this function
     *
     * @return a string */      
    public String getDescription() {
	return "Returns a square root of a number.";
    }
}

