import java.io.*;
import java.util.*;

/** This holds the data within each cell.
 */
public class Node {
    /** reserved */
    public static final int DEFAULT = 0;
    /** LETTERS+numbers: A1 */
    public static final int REL_ADDR = 1;
    /** $LETTERS$numbers: $A$1 */
    public static final int ABS_ADDR = 2;
    /** LETTERS: SUM */
    public static final int FUNCTION = 3;
    /** ( */
    public static final int LPAREN = 4;
    /** ) */
    public static final int RPAREN = 5;
    /**  all numbers and has .: 0.5 */
    public static final int NUMBER = 6;
    /** + - * / ^ */
    public static final int OPERATOR = 7;
    /** , */
    public static final int COMMA = 8;
    /** : */
    public static final int COLON = 9;
    /** an expression.  the exp field counts. */
    public static final int EXP = 10;
    
    // for each function param, its type is EXP
    
    private static final String[] desc = {
        "Default", "Rel_Addr", "Abs_Addr", "Function",
        "LBracket", "RBracket", "Number", "Operator",
        "Comma", "Colon", "Param" };
        
        private int type;  // which type the node is (see above 10 types)
        private String data; // the raw data
        private float number; // the numeric value
        private int row;
        private int col;
        
        private LinkedList exp; // a LinkedList for a function's params
        
    /*
     * Used for Function address range parameter (ADDR1:ADDR2)
     * An address range is ultimately represented as follows:
     *
     * node type: COLON
     * node.nextRange points to the start address (a node of REL_ADDR or
     * ABS_ADDR), the start address' nextRange points to the end address.
     */
        private Node nextRange;
        
        private boolean pending; // used for processing functions, see Formula
        
        /**
         * This is an empty node constructor
         */
        Node() {
        }
        
        /**
         * Node constructor
         *
         * @param node
         */
        Node(Node node) {
            type = node.type;
            if (data != null)
                data = new String(node.data);
            number = node.number;
            row = node.row;
            col = node.col;
        }
        
        /** get/set functions
         * @return the type from the list of 10 above
         */
        public int getType() { return type; }
        /** Determines if the types are the same
         * @param type the input type
         * @return true is the types are the same, false otherwise
         */
        public boolean isType(int type) { return this.type == type; }
        
        /** Gets the value of the cell.
         *
         * @return the value
         */
        public float getNumber() { return number; }
        /** Gets the data value in the cell.
         * @return the value
         */
        public String getData() { return data; }
        
        /** List the parameters.
         * @return the list of parameters
         */
        public LinkedList getParams() { return exp; }
        /** This function returns the Node that is set as
         * the next range in the range of this node.
         * @return  - a Node object.
         */
        public Node getNextRange() { return nextRange; }
        /** This function returns the paramater list used for
         * evaluating the formula contained in this node.
         * @return  - a LinkedList containing the parameters
         *   for the formula.
         */
        public LinkedList getExp() { return exp; }
        /** This function returns the row of this node.
         * @return  - an integer representing the row.
         */
        public int getRow() { return row; }
        /** This function returns the column of this node.
         * @return  - an integer representing the column.
         */
        public int getCol() { return col; }
        /** This function sets the type of this Node.
         * @param type  - an integer representing the type of
         *   data to becontained in this node.
         */
        public void setType(int type) { this.type = type; }
        /** This function sets the number data variable for
         * this node.
         * @param number  - a float number to be set.
         */
        public void setNumber(float number) { this.number = number; }
        /** This function sets the raw data variable for this node.
         * @param data  - a string of raw data to be set.
         */
        public void setData(String data) { this.data = data; }
        /** This function appends a character of raw data to the
         * end of the raw data variable.
         * @param data  - a character of raw data to be appended.
         */
        public void appendData(char data) { this.data += data; }
        /** This function appends a String of data to the end
         * of the raw data variable.
         * @param data  - a String of data to be appended.
         */
        public void appendData(String data) { this.data += data; }
        /** Sets the parameter list for the formula in this node.
         * Works in the same way as setExp function.
         * @param list  - a LinkedList of paramaters for the formula.
         */
        public void setParams(LinkedList list) { exp = list; }
        /** This function adds parameters to the paramater list
         * needed for the Formula in this function.
         * Adds to the "exp" variable (LinkedList).
         * @param node  - a Node object representing the paramater
         */
        public void addParam(Node node) {
            if (node.getExp().size()>0)
                exp.add(node);
        }
        /** This function sets the next node in the range
         * for this node.
         * @param node  - the next Node in the Range.
         */
        public void setNextRange(Node node) { nextRange = node; }
        /** This function sets the list of parameters needed
         * for the function thats in this node.
         * @param exp  - a LinkedList of all parameters needed
         *   for the formula in this node.
         */
        public void setExp(LinkedList exp) { this.exp = exp; }
        /** This function sets the pending status of this node.
         *
         * @param pending  - passed in a boolean value
         *   true - set status as pending
         *   false - set status as not pending
         */
        public void setPending(boolean pending) { this.pending = pending; }
        /** This function returns whether a node has a formula
         * thats currently pending evaluation.
         * @return  - a boolean value is returned.
         *   true - is pending
         *   false - is not pending
         */
        public boolean isPending() { return pending; }
        /** This function sets the row for this node.
         * @param row  - integer value for row.
         */
        public void setRow(int row) { this.row = row; }
        /** This function sets the column value for
         * this node.
         * @param col  - integer value for column.
         */
        public void setCol(int col) { this.col = col; }
        /** This function returns whether it has a value
         * the current node or not.
         * @return  - Returns a boolean
         *   true - Has value
         *   false - does not have value
         */
        public boolean hasValue() {
            return (type == Node.NUMBER || type == Node.REL_ADDR ||
            type == Node.ABS_ADDR || type == Node.FUNCTION ||
            type == Node.RPAREN);
        }
        
        /**  This string funciton is for debug purpose only.
         *  It converts the contents and attributes of
         *  the node into a string representation.
         *  (Node -> String)
         * @return  - This function returns a String object
         */
        public String toString() {
            //	return String.valueOf(type);
            
            switch (type) {
                case Node.NUMBER:
                    return Float.toString(number);
                case Node.REL_ADDR:
                    StringBuffer buf1 = new StringBuffer();
                    buf1.append('(');
                    buf1.append(String.valueOf(col));
                    buf1.append(',');
                    buf1.append(row);
                    buf1.append(')');
                    return buf1.toString();
                case Node.ABS_ADDR:
                    StringBuffer buf2 = new StringBuffer();
                    buf2.append('$');
                    buf2.append(translateColumn(col));
                    buf2.append('$');
                    buf2.append(translateRow(row));
                    return buf2.toString();
                case Node.FUNCTION:
                    return data+exp.toString();
                case Node.EXP:
                    return exp.toString();
                case Node.LPAREN:
                    return "(";
                case Node.RPAREN:
                    return ")";
                case Node.COMMA:
                    return ",";
                case Node.COLON:
                    return nextRange.toString()+":"+nextRange.nextRange.toString();
                    
                default:
                    return data;
            }
            
        }
        
        /** Given integer row and column, this function returns
         * a cellpoint object. (Node -> CellPoint)
         * @param row  - passed in as integer.
         * @param col - passed in a integer.
         * @return  - Returns the row and column in a CellPoint object.
         */
        public CellPoint toCellPoint(int row, int col) {
            if (isType(Node.REL_ADDR))
                return new CellPoint(getRow()+row, getCol()+col);
            else
                return new CellPoint(getRow(), getCol());
        }
        
        /**
         * This translates the string form of row into row number ('12' -> 12).
         *
         * @param row the string value of the row to be converted
         * @return the int value of the row
         */
        public static int translateRow(String row) {
            int r = Integer.parseInt(row);
            return r + UMSpreadSheet.m_baseRow - 1;
        }
        
        /**
         * This translates the int value of row into a string (12 -> '12').
         *
         * @param row the int value of the row to be converted
         * @return the string value of the row
         */
        public static String translateRow(int row) {
            if (row < 0)
                return null;
            else
                return String.valueOf(row+1-UMSpreadSheet.m_baseRow);
        }
        
        /**
         * This translates the string form of column into column number ('A' -> 1)
         *
         * @param column the string value of the column to be converted
         * @return the int value of the column
         */
        public static int translateColumn(String column) {
            int col = 0;
            for (int i = 0; i < column.length(); i++) {
                col  = col * 26 + (column.charAt(i) - 'A' + 1);
            }
            return col + UMSpreadSheet.m_baseCol -1;
        }
        
        /**
         * This translates the int form of column into column string (1 -> 'A')
         *
         * @param column the int value of the column to be converted
         * @return the string value of the column
         */
        public static String translateColumn(int column) {
            column = column - UMSpreadSheet.m_baseCol +1;
            if (column < 1)
                return null;
            String colstr = "";
            int div = 1;
            while (div > 0) {
                div = (column - 1)/26;
                colstr = (char)('A'+(column - div*26 - 1)) + colstr;
                column = div;
            }
            return colstr;
        }
        
        
        private static int min(int m, int n) {
            return m<n ? m:n;
        }
        
        private static int max(int m, int n) {
            return m>n ? m:n;
        }
        
        /**
         * Given one Node, this returns two CellPoints that identify the
         * upleft and downright cells.
         *
         * row and col matter because relative addresses need them.
         *
         * @param row the int value fo the row
         * @param col the int value of the column
         * @return two CellPoint value for the range
         */
        public CellPoint[] getAddressRange(int row, int col) {
            
            CellPoint[] addr = new CellPoint[2];
            if (isType(Node.COLON)) {
                // if it's an address range
                // get the start address
                CellPoint addr1 = getNextRange().toCellPoint(row, col);
                // get the end address
                CellPoint addr2 = getNextRange().getNextRange().
                toCellPoint(row, col);
                
                addr[0] = new CellPoint(
                min(addr1.getRow(), addr2.getRow()),
                min(addr1.getCol(), addr2.getCol())
                );
                addr[1] = new CellPoint(
                max(addr1.getRow(), addr2.getRow()),
                max(addr1.getCol(), addr2.getCol())
                );
            }
            else {
                // otherwise it's a normal address
                CellPoint cell = toCellPoint(row, col);
                addr[0] = cell;
                addr[1] = cell;
            }
            
            return addr;
        }
        
}
