
import java.awt.*;
import java.io.*;
import java.net.*;
import javax.swing.*;
import java.awt.datatransfer.*;
import java.util.*;
import java.awt.image.*;

/**
 * A Transferable which implements the capability required to transfer a
 * String.
 *
 * This Transferable properly supports <code>DataFlavor.stringFlavor</code>
 * and all equivalent flavors. Support for <code>DataFlavor.plainTextFlavor
 * </code> and all equivalent flavors is <b>deprecated</b>. No other
 * DataFlavors are supported.
 *
 * @see java.awt.datatransfer.DataFlavor.stringFlavor
 * @see java.awt.datatransfer.DataFlavor.plainTextFlavor
 */
public class UMClipObj implements Transferable, ClipboardOwner
 {

    private static final DataFlavor[] flavors = {
        DataFlavor.stringFlavor,
	DataFlavor.plainTextFlavor,
        DataFlavor.imageFlavor
    };

    private String data;
    private String m_text;
    private Image image;
    
    /**
     * Creates a Transferable capable of transferring the specified String.
     */
    UMClipObj(String simpleText, String fullText){

         this.data = fullText;
         this.m_text = simpleText;    
    }
    
    /**
     * Creates a Transferable capable of transferring a BufferedImage
     */
    UMClipObj(Image imageToClipboard){
        this.image = imageToClipboard;
    }

    /**
     * Returns an array of flavors in which this Transferable can provide
     * the data. <code>DataFlavor.stringFlavor</code> is properly supported.
     * Support for <code>DataFlavor.plainTextFlavor</code> is
     * <b>deprecated</b>.
     *
     * @return an array of length two, whose elements are <code>DataFlavor.
     *         stringFlavor</code> and <code>DataFlavor.plainTextFlavor</code>.
     */
    public DataFlavor[] getTransferDataFlavors() {
        // returning flavors itself would allow client code to modify
        // our internal behavior
	return (DataFlavor[])flavors.clone();
    }

    /** Returns whether the requested flavor is supported by this Transferable.
     *
     * @return true if <code>flavor</code> is equal to
     *   <code>DataFlavor.stringFlavor</code> or
     *   <code>DataFlavor.plainTextFlavor</code>; false if <code>flavor</code>
     *   is not one of the above flavors
     * @param flavor the requested flavor for the data */
    public boolean isDataFlavorSupported(DataFlavor flavor) {
        for (int i = 0; i < flavors.length; i++) {
	    if (flavor.equals(flavors[i])) {
	        return true;
	    }
	}
	return false;
    }
    
    /** Returns the Transferable's data in the requested DataFlavor if
     * possible. If the desired flavor is <code>DataFlavor.stringFlavor</code>,
     * or an equivalent flavor, the String representing the selection is
     * returned. If the desired flavor is </code>DataFlavor.plainTextFlavor
     * </code>, or an equivalent flavor, a Reader is returned. <b>Note:<b>
     * The behavior of this method for </code>DataFlavor.plainTextFlavor</code>
     * and equivalent DataFlavors is inconsistent with the definition of
     * <code>DataFlavor.plainTextFlavor</code>.
     *
     * @return the data in the requested flavor, as outlined above.
     * @see java.io.Reader#
     * @param flavor the requested flavor for the data
     * @throws UnsupportedFlavorException if the requested data flavor is
     *         not equivalent to either <code>DataFlavor.stringFlavor</code>
     *         or <code>DataFlavor.plainTextFlavor</code>.
     * @throws IOException if an IOException occurs while retrieving the data.
     *         By default, StringSelection never throws this exception, but a
     *         subclass may. */
    public Object getTransferData(DataFlavor flavor)
        throws UnsupportedFlavorException, IOException
    {
	if (flavor.equals(DataFlavor.stringFlavor)) {
            // returns clipboard data without font formats
            // used when pasting from other applications into TerpSpreadsheet (and vice versa)
            return (Object)m_text;
            
	} else if (flavor.equals(DataFlavor.plainTextFlavor)) {
            // returns copied data with font formats attached
            // used when pasting within TerpSpreadSheet
            return (Object)data;
        } else if (flavor.equals(DataFlavor.imageFlavor)){
            return (Object)image;
	} else {
            System.out.println("Exception: Attempting to access clipboard with unsupported DataFlavor");
	    throw new UnsupportedFlavorException(flavor);
	}
    }
    
    /** Method required for implementing clipboard.
     *
     * @param clipboard current clipboard
     * @param transferable ??
     */    
    public void lostOwnership(Clipboard clipboard, 
                              Transferable transferable) 
    { 
        System.out.println ("Lost Clipboard ownership"); 
    } 
} 




